"""Details of the Gwyddion Simple Field format."""

from typing import Final

import numpy as np

null_byte: Final[bytes] = b"\x00"
null_char: Final[str] = null_byte.decode()
gsf_magic_line: Final = "Gwyddion Simple Field 1.0\n"
# This cries for a TypedDict, but I have yet to find a sane way to specify "extra"
# (not required, not known in advance) keys.
gsf_known_metadata_types: Final = {
    "XRes": int,
    "YRes": int,
    "XReal": float,
    "YReal": float,
    "XOffset": float,
    "YOffset": float,
    "Title": str,
    "XYUnits": str,
    "ZUnits": str,
}
gsf_known_metadata_order: Final = {
    key: i for i, key in enumerate(gsf_known_metadata_types, start=1)
}
# 32-bit (4-bytes) little-endian floats.
gsf_dtype: Final = np.dtype("<f4")
# Row-major order.
gsf_array_order: Final = "C"


def gsf_padding_lenght(header_length: int) -> int:
    """Length of the null-byte padding between metadata and data."""
    return 4 - header_length % 4



"""Read Gwyddion Simple Field files."""

import logging
from contextlib import suppress
from pathlib import Path
from typing import Any, cast

import numpy as np
from numpy.typing import NDArray

from .format import (
    gsf_dtype,
    gsf_known_metadata_types,
    gsf_magic_line,
    gsf_padding_lenght,
    null_byte,
)

log = logging.getLogger(__name__)


def read_gsf(
    path: Path | str,
) -> tuple[NDArray[np.float32], dict[str, Any]]:
    """Read a Gwyddion Simple Field file (.gsf).

    Parameters
    ----------
        path
            Path to the file to be read.

    Returns
    -------
        metadata
            A dict of metadata. The fields XRes and YRes are not included,
            since they would be a duplicate of data.shape. Custom fields not mentioned
            in the Gwyddion Simple Field specification are read as strings.
        data
            A 2-dimensional NumPy array of float32.

    Raises
    ------
        ValueError
            If the file to be read is not a Gwyddion Simple Field.
        KeyError
            If required metadata is missing from the file to be read.
    """
    path = Path(path)
    metadata = {}
    log.info("Reading %s", path)

    with path.open("rb") as file:
        # Check magic line.
        if file.readline().decode() != gsf_magic_line:
            msg = f"Magic line not found at the beginning of {path}"
            raise ValueError(msg)

        # Read metadata.
        # Peek does not do what you think it does.
        # https://stackoverflow.com/a/24474743
        while file.peek(1)[:1] != null_byte:
            key, value = file.readline().decode().split("=")
            metadata[key.strip()] = value.strip()
        for key, type_ in gsf_known_metadata_types.items():
            with suppress(KeyError):
                metadata[key] = type_(metadata[key])

        # Skip null-byte padding.
        header_length = file.tell()
        file.seek(gsf_padding_lenght(header_length), 1)

        # Read data.
        # cast is there to persuade mypy.
        shape = cast("int", metadata["YRes"]), cast("int", metadata["XRes"])
        data_size = gsf_dtype.itemsize * shape[0] * shape[1]
        data = np.frombuffer(file.read(data_size), dtype=gsf_dtype).reshape(shape)

        if file.read(1) != b"":
            msg = f"Unexpected additional data found at the end of {path}"
            raise ValueError(msg)

    log.info(
        "Read an image of width: %s px, height: %s px", data.shape[1], data.shape[0]
    )
    log.debug("Read metadata: %s", metadata)

    # Do not duplicate information already present in data.shape.
    del metadata["XRes"]
    del metadata["YRes"]

    return data, metadata



"""Test that what is read is what we expect, given what was written."""

from pathlib import Path
from typing import Any

import numpy as np
import pytest
from numpy.typing import NDArray

from gsffile import read_gsf, write_gsf
from gsffile.write import prepare_data, prepare_metadata

rng = np.random.default_rng(seed=1)


def assert_roundtrip_ok(
    tmp_path: Path, data: NDArray[np.float32], meta: dict[Any, Any] | None = None
) -> None:
    """Assert that what is read is what we expect, given what was written."""
    path = tmp_path / "test.gsf"
    write_gsf(path, data, meta)
    data_2, meta_2 = read_gsf(path)
    expected_data = prepare_data(data)
    expected_meta = prepare_metadata(meta)
    assert meta_2 == expected_meta
    # Check metadata order.
    assert list(meta_2.keys()) == list(expected_meta.keys())
    np.testing.assert_array_equal(data_2, expected_data)


allowed_shapes = [(3, 2), (3, 1), (1, 3), (3,), (1,), ()]


@pytest.mark.parametrize("shape", allowed_shapes, ids=lambda shape: f"shape {shape}")
def test_allowed_shapes(tmp_path: Path, shape: tuple[int]) -> None:
    """Test 0-, 1-, and 2-dimensional random data."""
    data = rng.uniform(size=shape).astype(np.float32)
    assert_roundtrip_ok(tmp_path, data)


def test_meta(tmp_path: Path) -> None:
    """Test metadata roundtrip."""
    # Optional fields of non-str type, such as XReal, would fail this test.
    data = np.zeros((2, 3), dtype=np.float32)
    meta = {
        "string": "a string",
        "float": np.sqrt(3),
        "int": -0,
        "  whitespace  ": " ",
        "multi-byte characters": "💣",
        "ZUnits": "m",
        "XYUnits": "m",
    }
    assert_roundtrip_ok(tmp_path, data, meta)


def test_non_finite(tmp_path: Path) -> None:
    """Test roundtrip for inf and NaN."""
    data = np.array([[np.inf, -np.inf], [np.nan, -0.0]], dtype=np.float32)
    assert_roundtrip_ok(tmp_path, data)


def test_path_as_str(tmp_path: Path) -> None:
    """Test passing a str as path."""
    data = np.zeros((2, 3), dtype=np.float32)
    path = tmp_path / "test.gsf"
    write_gsf(str(path), data)
    _ = read_gsf(str(path))
